/*

  freedom - Gameboy cardtridge converter from liberty 
            format to palmboy

	v1.0   (c) 2001 by Till Harbaum
	v1.1   corrected bug by Bodo Wenzel
*/

#include <PalmOS.h>
#include <PalmCompatibility.h>

#include "freedom.h"
#include "axxPac.h"

#define DEBUG

#ifdef DEBUG
#define DEBUG_PORT 0x30000000
#define DEBUG_MSG(format, args...) { char tmp[128]; int xyz; StrPrintF(tmp, format, ## args); for(xyz=0;xyz<StrLen(tmp);xyz++) *((char*)DEBUG_PORT)=tmp[xyz]; }
#else
#define DEBUG_MSG(format, args...)
#endif

#define ROM_BANK       16384U
#define RAM_BANK       8192U
#define MAX_ROM_BANKS  256
#define MAX_RAM_BANKS  4
#define MAX_RAM_SIZE   (MAX_RAM_BANKS*RAM_BANK)

static MenuBarPtr CurrentMenu = NULL;

#define LIB_CREATOR 'Gmbt'
#define LIB_TYPE    'gROM'
#define LIB_RSRC    '_dat'

#define PB_CREATOR  'PBoy'
#define PB_TYPE     'GAME'

typedef struct {
  char    name[32];
  UInt16  card;
  LocalID dbid;
  UInt16  size;
  long    ram;
} db_info;

db_info *db = NULL;
int num_db=0;
int scl_offset=0;

UInt LibRef = 0;

typedef  struct {
  unsigned char  unused1[0x134];
  char           name[17];
  unsigned char  unused2[2];
  unsigned char  ctype;
  unsigned char  romsz;
  unsigned char  ramsz;
}  CARTR_HEADER;

void *GetObjectPtr(FormPtr frm, Word object) {
  return FrmGetObjectPtr(frm, FrmGetObjectIndex (frm, object));
}

static Err  BeamBytes(ExgSocketPtr s, void *buffer, ULong bytesToSend) {
   Err err = 0;
   
   while (!err && bytesToSend > 0) {
      ULong bytesSent = ExgSend(s, buffer, bytesToSend, &err);
      bytesToSend -= bytesSent;
      buffer = ((char *) buffer) + bytesSent;
   }
   return err;
}

Err
write_proc(const void *data, ULong *size, void *udata) {
  /* send data ... */
  return BeamBytes((ExgSocketPtr)udata, (void*)data, *size);
}

Boolean db_beam(char *name) {
  ExgSocketType  s;
  Err      err;

  MemSet(&s, sizeof(s), 0);
  s.description = name;
  s.target = 'lnch';
  s.length = 1;
  
  if (!ExgPut(&s)) {
    /* beam db */
    err = ExgDBWrite(write_proc, &s, s.description, 0, 0);
    err = ExgDisconnect(&s, err);
  }
  return err;
}

void launch_palmboy(void) {
  LocalID dbId;
  DWord result;
	
  /* delete database if already existant */
  if((dbId = DmFindDatabase(0, "PalmBoy"))!=0) {
    SysUIAppSwitch(0, dbId, sysAppLaunchCmdNormalLaunch, NULL);
  } else {
    FrmCustomAlert(alt_err, "Unable to locate PalmBoy!",0,0);
  }
}

static int scan_games(db_info **entries, ULong type, ULong creator);
void display_games(FormPtr frm);

long get_ram_size(CARTR_HEADER* gb_header) {
  long ram_size;

  switch (gb_header->ctype) {
  case 0x05:
  case 0x06:
    ram_size = 0x200L/2; /* just 512 nibble! */
    break;
  case 0xfe:
  case 0xff:
    ram_size = 0x8000L;
    break;
  default:
    ram_size = (gb_header->ramsz == 0) ? 0 : 0x200L << (2 * gb_header->ramsz);
    break;
  }
  
  if (ram_size > MAX_RAM_SIZE) {
    FrmCustomAlert(alt_err, "Too much RAM in this cartridge!", 0,0);
    return -1;
  }

  return ram_size;
}

typedef  struct {
  int   id;
  int   mcb;
  char  *name;
} CARTR_TYPE;

static  CARTR_TYPE  known_types[]={
  { 0x00, 0,"ROM ONLY" },
  { 0x01, 1,"MBC1" },
  { 0x02, 1,"MBC1+RAM" },
  { 0x03, 1,"MBC1+RAM+BAT" },
  { 0x05, 2,"MBC2" },
  { 0x06, 2,"MBC2+BAT" },
  { 0x08, 0,"RAM" },
  { 0x09, 0,"RAM+BAT" },
  { 0x0B,-1,"MMM01" },
  { 0x0C,-1,"MMM01+RAM" },
  { 0x0D,-1,"MMM01+RAM+BAT" },
  { 0x0F, 3,"MBC3+BAT+RTC" },
  { 0x10, 3,"MBC3+RAM+BAT+RTC" },
  { 0x11, 3,"MBC3" },
  { 0x12, 3,"MBC3+RAM" },
  { 0x13, 3,"MBC3+RAM+BAT" },
  { 0x15,-1,"MBC4" },
  { 0x16,-1,"MBC4+RAM" },
  { 0x17,-1,"MBC4+RAM+BAT" },
  { 0x19, 5,"MBC5" },
  { 0x1A, 5,"MBC5+RAM" },
  { 0x1B, 5,"MBC5+RAM+BAT" },
  { 0x1C, 5,"MBC5+MOT" },
  { 0x1D, 5,"MBC5+RAM+MOT" },
  { 0x1E, 5,"MBC5+RAM+BAT+MOT" },
  { 0xFC,-1,"Nintendo Pocket Camera" },
  { 0xFD,-1,"Bandai TAMA5" },
  { 0xFE,-1,"Hudson HuC-3" },
  { 0xFF,-1,"Hudson HuC-1" },
  { -1,  -1,"Unknown" }
};

Boolean isprint(char ch) {
  return((ch>=32)&&(ch<127));
}

char fname[AXXPAC_FSEL_BUFFERSIZE] = "";

Boolean axxpac_import(void) {
  axxPacFD infile=-1;
  CARTR_HEADER *gb_header;

  int type, i;
  int rom_banks, ram_banks;
  long ram_size;
  LocalID dst_dbId;
  DmOpenRef dst=0;
  Boolean ret = false;
  char *buffer = NULL;
  UInt at;
  VoidHand rsc;
  MemPtr dst_ptr;
  char name[32];

  if(LibRef == NULL) {
    FrmCustomAlert(alt_err, 
       "No axxPac installed. See www.axxpac.de for details!",0,0);
    goto axxpac_end;
  }

  i = axxPacFsel(LibRef, "Select gameboy cartridge", fname, "*", false);
  /* file error */
  if(i<0) {
    FrmCustomAlert(alt_err, axxPacStrError(LibRef), 0, 0);
    goto axxpac_end;
  }
  
  /* cancelled */
  if(i == 0) {
    ret = true;
    goto axxpac_end;
  }
  
  /* get some buffer space for the ROM chunks */
  if((gb_header = (CARTR_HEADER*)buffer = MemPtrNew(32768)) == NULL) {
    FrmCustomAlert(alt_err, "Out of internal memory!", 0,0);
    goto axxpac_end;
  }

  /* open file */
  if((infile = axxPacOpen(LibRef, fname, O_RdOnly))<0) {
    FrmCustomAlert(alt_err, "Unable to open file: ^1", 
		   axxPacStrError(LibRef), 0);
    goto axxpac_end;
  }
  
  /* scan some image values */
  if (axxPacRead(LibRef, infile, buffer, 32768) != 32768) {
    FrmCustomAlert(alt_err, "Can't read input file ^1!", fname,0);
    goto axxpac_end;
  }

  /* save cart name */
  StrCopy(name, gb_header->name);
  
  /* search for cartridge type */
  for (type = 0; known_types[type].id >= 0 &&
	 known_types[type].id != gb_header->ctype; type++);
  
  if (known_types[type].mcb < 0) {
    FrmCustomAlert(alt_err, "Unsupported cartridge type ^1!",
		   known_types[type].name, 0);
    goto axxpac_end;
  }
  
  /* determine cartridge size */
  rom_banks = 2 << (gb_header->romsz & 0x0f);
  if ((gb_header->romsz & 0xf0) != 0)
    rom_banks += 2 << (gb_header->romsz >> 4);
  
  if(rom_banks > MAX_ROM_BANKS) {
    FrmCustomAlert(alt_err, "Too much ROM in this cartridge!",0,0);
    goto axxpac_end;
  }
  
  if((ram_size = get_ram_size(gb_header))<0) {
    FrmCustomAlert(alt_err, "Too much RAM in this cartridge!",0,0);
    goto axxpac_end;
  }
  
  if (ram_size == 0)  ram_banks = 0;
  else                ram_banks = (int)((ram_size - 1) / RAM_BANK) + 1;
  
  /* clean up cartridge name */
  if (gb_header->name[15] == '\xc0') {
    FrmCustomAlert(alt_err, "Warning: color-only game!",0,0);
    for (i = 0; i < 11; i++) {
      if (gb_header->name[i] == '\0')
	break;
      if (!isprint(gb_header->name[i]))
	gb_header->name[i] = ' ';
    }
  } else if (gb_header->name[15] == '\x80') {
    /* game with color and gray support */
    for (i = 0; i < 15; i++) {
      if (gb_header->name[i] == '\0')
	break;
      if (!isprint(gb_header->name[i]))
	gb_header->name[i] = ' ';
    }
  } else {
    /* old game, gray-only */
    for (i = 0; i < 16; i++) {
      if (gb_header->name[i] == '\0')
	break;
      if (!isprint(gb_header->name[i]))
	gb_header->name[i] = ' ';
    }
  }
  
  gb_header->name[i] = '\0';
  
  /* create a new database */
  
  /* database exists? */
  if((dst_dbId = DmFindDatabase(0, gb_header->name))!=0) {
    if(FrmAlert(alt_overwrite)==0) {
      if(DmDeleteDatabase(0, dst_dbId) != 0) {
	FrmCustomAlert(alt_err, "Unable to delete existing database!",0,0);
	goto axxpac_end;
      }
    } else
      goto axxpac_end;
  }
  
  /* create destination database */
  if(DmCreateDatabase(0, gb_header->name, PB_CREATOR, PB_TYPE, false)!=0) {
    FrmCustomAlert(alt_err, "Unable to create destination database!",0,0);
    goto axxpac_end;
  }
  
  /* find destination database */
  if((dst_dbId = DmFindDatabase(0, gb_header->name))==0) {
    FrmCustomAlert(alt_err, "Unable to find destination database!",0,0);
    goto axxpac_end;
  }
  
  /* open destination database */
  if((dst = DmOpenDatabase(0, dst_dbId, dmModeReadWrite))==0) {
    FrmCustomAlert(alt_err, "Unable to open destination database!",0,0);
    goto axxpac_end;
  }
  
  /* ceate all ROM records */
  
  for(i=1;i<rom_banks;i++) {
    
    /* create record */
    at = i-1;
    
    if((rsc = DmNewRecord(dst, &at, (i==1)?32768:16384))==NULL) {
      FrmCustomAlert(alt_err, "Out of memory creating ROM record.",0,0);
      goto axxpac_end;
    }
    
    /* get address of new record */
    dst_ptr = MemHandleLock(rsc);
    
    /* copy data to record */
    DmWrite(dst_ptr, 0, buffer, (i==1)?32768:16384);
      
    /* unlock source and destination ids */
    MemHandleUnlock(rsc);
    
    DmReleaseRecord(dst, at, true);
    
    /* read next ROM chunk */
    if(i!=rom_banks-1) {
      
      if (axxPacRead(LibRef, infile, buffer, 16384) != 16384) {
	FrmCustomAlert(alt_err, "Can't read input file!",0,0);
	goto axxpac_end;
      }
    }
  }
  
  /* free ROM buffer */
  MemPtrFree(buffer);
  buffer = NULL;
  
  if(ram_size != 0) {
    long ram_record_size;

    if (ram_size < RAM_BANK)
      ram_record_size = RAM_BANK;
    else
      ram_record_size = ram_size;

    /* get RAM buffer */
    if((buffer = MemPtrNew(ram_record_size)) == NULL) {
      FrmCustomAlert(alt_err, "Out of internal memory!", 0,0);
	goto axxpac_end;
    }
    
    /* erase RAM */
    MemSet(buffer, ram_record_size, 0);
    
    /* ram at next record */
    at += 1;
    if((rsc = DmNewRecord(dst, &at, ram_record_size))==NULL) {
      FrmCustomAlert(alt_err, "Out of memory creating RAM record.",0,0);
      goto axxpac_end;
    }
    
    /* get address of new record */
    dst_ptr = MemHandleLock(rsc);
    
    DmWrite(dst_ptr, 0, buffer, ram_record_size);
    
    MemHandleUnlock(rsc);
    
    DmReleaseRecord(dst, at, true);
      
    MemPtrFree(buffer);
    buffer=NULL;
  }
  
  ret = true;

#if 0
  FrmCustomAlert(alt_success, name, 0, 0);
#endif
  
 axxpac_end:
  /* free memory */
  if(buffer != NULL) MemPtrFree(buffer);
  
  /* close database */
  if(dst != 0) DmCloseDatabase(dst);
  
  /* close file */
  if(infile >= 0)  axxPacClose(LibRef, infile);
    
  if(ret) {
    num_db = scan_games(&db, PB_TYPE, PB_CREATOR);
    display_games(FrmGetActiveForm());
  } else {
    /* delete database on error */
    if((dst_dbId = DmFindDatabase(0, name))!=0) {
      DmDeleteDatabase(0, dst_dbId);
    }
  }

  return ret;
}

Boolean convert_db(char *name) {
  DmOpenRef src=0, dst=0;
  LocalID src_dbId, dst_dbId, src_ch;
  UInt16 entries, i;
  MemPtr src_ptr, dst_ptr;
  Boolean ret=false;
  UInt at;
  VoidHand rsc;
  CARTR_HEADER gb_header;
  long ram_size;
  int type;
  char *ram, *pb_name;

  /* check if the first five characters of the name are "Gmbt_" */
  if(StrNCompare("Gmbt_", name, 5) != 0) {
    FrmCustomAlert(alt_err, "Illegal database name!",0,0);
    goto conv_end;
  }

  /* find database */
  if((src_dbId = DmFindDatabase(0, name))==0) {
    FrmCustomAlert(alt_err, "Unable to find source database!",0,0);
    goto conv_end;
  }

  /* open source database */
  if((src = DmOpenDatabase(0, src_dbId, dmModeReadOnly))==0) {
    FrmCustomAlert(alt_err, "Unable to open source database!",0,0);
    goto conv_end;
  }
 
  /* skip the first five characters for palmboy name */
  pb_name = name + 5;

  /* database exists? */
  if((dst_dbId = DmFindDatabase(0, pb_name))!=0) {
    if(FrmAlert(alt_overwrite)==0) {
      if(DmDeleteDatabase(0, dst_dbId) != 0) {
	FrmCustomAlert(alt_err, "Unable to delete existing database!",0,0);
	goto conv_end;
      }
    } else
      goto conv_end;
  }

  /* create destination database */
  if(DmCreateDatabase(0, pb_name, PB_CREATOR, PB_TYPE, false)!=0) {
    FrmCustomAlert(alt_err, "Unable to create destination database!",0,0);
    goto conv_end;
  }
 
  /* find destination database */
  if((dst_dbId = DmFindDatabase(0, pb_name))==0) {
    FrmCustomAlert(alt_err, "Unable to find destination database!",0,0);
    goto conv_end;
  }

  /* open destination database */
  if((dst = DmOpenDatabase(0, dst_dbId, dmModeReadWrite))==0) {
    FrmCustomAlert(alt_err, "Unable to open destination database!",0,0);
    goto conv_end;
  }

  /* determine number of source records */
  entries = DmNumResources(src);

  for(i=0;i<entries;i++) {
    /* get size of resources */
    if(DmResourceInfo(src, i, NULL, NULL, &src_ch)!=0) {
      FrmCustomAlert(alt_err, "Unable to get resource info.",0,0);
      goto conv_end;
    }

    /* create new records */
    if(i != 1) {
      /* create a new record in palmboy database */
      at = (i==0)?0:i-1;

      if((rsc = DmNewRecord(dst, &at, (i==0)?32768:16384))==NULL) {
	FrmCustomAlert(alt_err, "Out of memory while creating ROM record.",0,0);
	goto conv_end;
      }
    } else
      DmGetRecord(dst, at);

    /* get address of new record */
    dst_ptr = MemHandleLock(rsc);

    src_ptr = MemLocalIDToLockedPtr(src_ch, 0);

    /* save cartridge info */
    if(i==0)
      MemMove(&gb_header, src_ptr, sizeof(CARTR_HEADER));

    /* copy data to record */
    if(i != 1) {
      DmWrite(dst_ptr, 0, src_ptr, 16384);
    } else {
      DmWrite(dst_ptr, 16384, src_ptr, 16384);      
    }

    /* unlock source and destination ids */
    MemHandleUnlock(rsc);
    MemPtrUnlock(src_ptr);

    DmReleaseRecord(dst, at, true);
  }

  /* check cartridge type */
  for (type = 0; known_types[type].id >= 0 &&
                 known_types[type].id != gb_header.ctype; type++);

  if (known_types[type].mcb < 0) {
    FrmCustomAlert(alt_err, "Unsupported cartridge type!",0,0);
    DmCloseDatabase(dst);
    
    if((dst_dbId = DmFindDatabase(0, pb_name))!=0)
      DmDeleteDatabase(0, dst_dbId);
      
    return false;
  }

  /* now attach RAM contents */
  /* determine RAM size */
  if((ram_size = get_ram_size(&gb_header)) < 0)
    goto conv_end;

  if(ram_size != 0) {
    long ram_record_size;

    if (ram_size < RAM_BANK)
      ram_record_size = RAM_BANK;
    else
      ram_record_size = ram_size;

    /* get RAM buffer */
    if((ram = MemPtrNew(ram_record_size)) == NULL) {
      FrmCustomAlert(alt_err, "Out of internal memory!", 0,0);
      goto conv_end;
    }

    /* erase RAM */
    MemSet(ram, ram_record_size, 0);
    
    /* ram at next record */
    at += 1;
    if((rsc = DmNewRecord(dst, &at, ram_record_size))==NULL) {
      FrmCustomAlert(alt_err, "Out of memory creating RAM record.",0,0);
      goto conv_end;
    }

    /* get address of new record */
    dst_ptr = MemHandleLock(rsc);

    DmWrite(dst_ptr, 0, ram, ram_record_size);
    
    MemHandleUnlock(rsc);
    
    DmReleaseRecord(dst, at, true);
    
    MemPtrFree(ram);
  }

  /* everything's fine! */
  ret = true;

#if 0
  FrmCustomAlert(alt_success, gb_header.name, 0,0);
#endif

 conv_end:
  /* close databases */
  if(src != 0) DmCloseDatabase(src);
  if(dst != 0) DmCloseDatabase(dst);

  if(!ret) {
    /* delete database on error */
    if((dst_dbId = DmFindDatabase(0, gb_header.name))!=0) {
      DmDeleteDatabase(0, dst_dbId);
    }
  }

  return ret;
}

/* write length limited string followed by dots at pos */
static void 
win_draw_chars_len(int x, int y, SWord max, char *str) {
  SWord lenp;
  Boolean fit;
  const char dots[]="...";

  lenp = StrLen(str);
  fit = true;
  FntCharsInWidth(str, &max, &lenp, &fit);

  /* something cut off? */
  if(lenp != StrLen(str)) {
    max = max-FntLineWidth((char*)dots,3);
    FntCharsInWidth(str, &max, &lenp, &fit);
    WinDrawChars(str, lenp, x, y);
    WinDrawChars((char*)dots, 3, x+max, y);
  } else
    WinDrawChars(str, lenp, x, y);
}

static Boolean LibFormHandleEvent(EventPtr event) {
  return false;
}

#define XLINES 10
void draw_lib(FormPtr frm, db_info *entry, int i, int offset) {
  RectangleType rect1 = {{11,21},{127,XLINES*11}};
  RectangleType rect2 = {{10,20},{136,XLINES*11}};
  int k,j = (i>XLINES)?XLINES:i;

  /* erase file area */
  WinEraseRectangle(&rect1, 0);

  /* draw rectangle around file area */
  WinDrawRectangleFrame(1, &rect2);

  for(k=0;k<j;k++) {
    win_draw_chars_len(12, 20+11*k, 130, entry[k+offset].name);

    /* enable button for this entry */
    CtlSetEnabled(GetObjectPtr(frm, rsc_name0 + k), true);
  }

  /* disable remaining buttons */
  while(k<XLINES) {
    /* enable button for this entry */
    CtlSetEnabled(GetObjectPtr(frm, rsc_name0 + k++), false);
  }
}

void liberty_import(void) {
  FormPtr frm;
  Boolean quit=false;
  EventType event;
  int offset=0, libs=0;
  db_info *ldb = NULL;

  if((libs = scan_games(&ldb, LIB_TYPE, LIB_CREATOR))<=0) {
    FrmCustomAlert(alt_err, "No Liberty cartridge images installed!",0,0);
    return;
  }

  frm = FrmInitForm(rsc_fsel);
  FrmSetActiveForm(frm);
  FrmSetEventHandler(frm, LibFormHandleEvent);
  FrmPopupForm(rsc_fsel);
  FrmDrawForm(frm);

  draw_lib(frm, ldb, libs, offset);

  do {
    /* exit loop every ten system ticks (about 10Hz) */
    EvtGetEvent(&event, evtWaitForever);  

    if (!SysHandleEvent(&event)) {
      switch (event.eType) {

      case sclRepeatEvent:
	offset = event.data.sclExit.newValue;
	draw_lib(frm, ldb, libs, offset);
	break;
	
      case ctlSelectEvent:
	switch(event.data.ctlEnter.controlID) {
	case rsc_name0:
	case rsc_name1:
	case rsc_name2:
	case rsc_name3:
	case rsc_name4:
	case rsc_name5:
	case rsc_name6:
	case rsc_name7:
	case rsc_name8:
	case rsc_name9:
	  convert_db(ldb[event.data.ctlEnter.controlID-rsc_name0+offset].name);
	  quit = true;
	  break;
	  
	case rsc_cancel:
	  quit = true;
	  break;
	}
	break;
      }
      FrmDispatchEvent(&event);
    }

    if((event.eType) == appStopEvent) {
      /* put event back into queue */
      EvtAddEventToQueue(&event);
      quit=true;
    }
 } while (!quit);

  FrmReturnToForm(0);

  num_db = scan_games(&db, PB_TYPE, PB_CREATOR);
  display_games(FrmGetActiveForm());

fs_quit:
  if(ldb != NULL) MemPtrFree(ldb);

  return;
}


#define LINES 10

void draw_entries(FormPtr frm, db_info *entry, int i, int offset) {
  RectangleType rect = {{10,32},{128,LINES*11}};
  int k,j = (i>LINES)?LINES:i;
  char str[16];

  /* erase file area */
  WinEraseRectangle(&rect, 0);

  for(k=0;k<j;k++) {
    win_draw_chars_len(11, 32+11*k, 99, entry[k+offset].name);

    StrPrintF(str, "%d kB", entry[k+offset].size);
    WinDrawChars(str, StrLen(str), 137-FntLineWidth(str, StrLen(str)), 32+11*k);

    /* enable button for this entry */
    CtlSetEnabled(GetObjectPtr(frm, db_name0 + k), true);
  }

  /* disable remaining entries */
  while(k<LINES) {
    /* disable button for this entry */
    CtlSetEnabled(GetObjectPtr(frm, db_name0 + k++), false);
  }
}

static Int 
compare_entry(VoidPtr object1, VoidPtr object2, Long other) {
  db_info *d1=object1, *d2=object2;
  
  return StrCaselessCompare(d1->name, d2->name);
}

void display_games(FormPtr frm) {
  int i;
  ScrollBarPtr bar = GetObjectPtr(frm, scrollbar);

  /* update scrollbar */
  if(num_db>LINES) SclSetScrollBar(bar, 0, 0, num_db-LINES, LINES-1);
  else             SclSetScrollBar(bar, 0, 0, 0, 0);
  
  /* and draw them ... */
  draw_entries(frm, db, num_db, scl_offset);
}

static int scan_games(db_info **entries, ULong type, ULong creator) {
  DmSearchStateType stateInfo;
  Err err;
  LocalID dbid;
  UInt16 card, i, cnt_db;
  Boolean first;
  DmOpenRef Ref;
  CARTR_HEADER *gb_header;
  VoidHand hand;
  int rom_banks;

  if(*entries != NULL) {
    MemPtrFree(*entries);
    *entries = NULL;
  }
    
#if 0
  /* allocate buffer for all databases */
  if((*entries = MemPtrNew(sizeof(db_info) * 20)) == NULL) {
    FrmCustomAlert(alt_err, "Out of memory allocating database buffer.",0,0);
    return 0;
  }

  for(i=0;i<20;i++)
    StrPrintF((*entries)[i].name, "Name: %d", i);

  cnt_db = 20;

#else
  /* scan for number of level databases */
  cnt_db=0; first = true;

  do {
    err = DmGetNextDatabaseByTypeCreator(first, 
			 &stateInfo, type, 
			 creator, false, &card, &dbid);

    if(!err) cnt_db++;
    first = false;
  } while(!err);

  if(cnt_db==0) return 0;

  /* allocate buffer for all databases */
  if((*entries = MemPtrNew(sizeof(db_info) * (1+cnt_db))) == NULL) {
    FrmCustomAlert(alt_err, "Out of memory allocating database buffer.",0,0);
    return 0;
  }

  /* once again, but this time save the database infos */
  cnt_db=0; first = true;
  do {
    err = DmGetNextDatabaseByTypeCreator(first, 
	    &stateInfo, type, creator, false, 
            &((*entries)[cnt_db].card), &((*entries)[cnt_db].dbid));

    if(!err) cnt_db++;
    first = false;
  } while(!err);

  /* get db names */
  for(i=0;i<cnt_db;i++) {
    MemSet((*entries)[i].name, 32, 0);
    DmDatabaseInfo((*entries)[i].card, 
		   (*entries)[i].dbid, 
		   (*entries)[i].name, 
                   0, 0, 0, 0, 0, 0, 0, 0, 0, 0);

    /* determine database size (only for palmboy images) */
    if(creator == PB_CREATOR) {
      /* open destination database */
      if((Ref = DmOpenDatabase(0, (*entries)[i].dbid, dmModeReadOnly))!=0) {
	/* load first record */
	hand = DmGetRecord(Ref, 0);
	gb_header = MemHandleLock(hand);
	
	/* determine cartridge size */
	rom_banks = 2 << (gb_header->romsz & 0x0f);
	if ((gb_header->romsz & 0xf0) != 0)
	  rom_banks += 2 << (gb_header->romsz >> 4);
	
	(*entries)[i].size = 16 * rom_banks;
	(*entries)[i].ram  = get_ram_size(gb_header);
	
	MemHandleUnlock(hand);
	DmReleaseRecord(Ref, 0, false);
	
	DmCloseDatabase(Ref);
      }
    }
  }
#endif

  SysQSort( (Byte*)*entries, cnt_db, sizeof(db_info), compare_entry, 0);

  return cnt_db;
}

static Boolean StartApplication(void)
{
  Err err;

  /* try to open already installed axxPac driver lib */
  err = SysLibFind(AXXPAC_LIB_NAME, &LibRef);

  /* didn't work, try to load it */
  if (err) err = SysLibLoad('libr', AXXPAC_LIB_CREATOR, &LibRef);
    
  if(err) {
    /* ERROR: unable to access library */
    LibRef = 0;
  } else {
    /* try to open lib */
    if(axxPacLibOpen(LibRef) != 0) {
      /* ERROR: unable to open library */
      LibRef = 0;
    }
  }

  FrmGotoForm(frm_Main);
  return true;
}

static void StopApplication(void)
{
  Err err;
  UInt numapps;

  if(LibRef != 0) {
    err = axxPacLibClose(LibRef, &numapps);

    if(!err) {
      if (numapps == 0)
	SysLibRemove(LibRef);
    }
  }
}

void show_db_info(char *name, int rom, long ram) {
  int i;
  char str[2][16];
  LocalID dbId;

  StrPrintF(str[0], "%d kB", rom);

  if(ram <= 0)      StrCopy(str[1],   "no");
  else if(ram<1024) StrPrintF(str[1], "%ld Bytes", ram);
  else              StrPrintF(str[1], "%ld kB", ram/1024);

  i = FrmCustomAlert(alt_dbinfo, name, str[0], str[1]);

  if(i == 0) db_beam(name);

  if(i == 1) {
    if((dbId = DmFindDatabase(0, name))!=0) {
      if(FrmCustomAlert(alt_delete, name, 0,0)==0) {
	if(DmDeleteDatabase(0, dbId) != 0) {
	  FrmCustomAlert(alt_err, "Unable to delete existing database!",0,0);
	}
      }
    }
    num_db = scan_games(&db, PB_TYPE, PB_CREATOR);
    display_games(FrmGetActiveForm());
  }
}

/////////////////////////
// MainFormHandleEvent
/////////////////////////

static Boolean MainFormHandleEvent(EventPtr event)
{
  Boolean handled = false;
  FormPtr frm = FrmGetActiveForm();

  switch (event->eType) {

  case menuEvent:
    switch (event->data.menu.itemID) {

    case mnuAbout:
      FrmAlert(alt_about);
      handled = true;
      break;

    case mnuImport:
      axxpac_import();
      handled = true;
      break;

    case mnuLiberty:
      liberty_import();
      handled = true;
      break;

    case mnuPalmboy:
      launch_palmboy();
      handled = true;
      break;
    }
    break;

  case sclRepeatEvent:
    scl_offset = event->data.sclExit.newValue;
    draw_entries(frm, db, num_db, scl_offset);
    break;
	 
  case ctlSelectEvent:
    if((event->data.ctlEnter.controlID>=db_name0)&&
       (event->data.ctlEnter.controlID<=db_name9)) {
      show_db_info(db[event->data.ctlEnter.controlID-db_name0+scl_offset].name,
		   db[event->data.ctlEnter.controlID-db_name0+scl_offset].size,
		   db[event->data.ctlEnter.controlID-db_name0+scl_offset].ram);
      handled = true;
    }
    break;
    
  case frmOpenEvent:
    FrmDrawForm(frm);
    num_db = scan_games(&db, PB_TYPE, PB_CREATOR);
    display_games(frm);
    handled = true;
    break;
  }
    
  return(handled);
}

/////////////////////////
// ApplicationHandleEvent
/////////////////////////

static Boolean ApplicationHandleEvent(EventPtr event)
{
  FormPtr form;
  Int16    formId;
  Boolean handled = false;
  
  if (event->eType == frmLoadEvent) {
    formId = event->data.frmLoad.formID;
    form = FrmInitForm(formId);
    FrmSetActiveForm(form);
    
    if(formId ==frm_Main)
      FrmSetEventHandler(form, MainFormHandleEvent);

    handled = true;
  }
  
  return handled;
}

/////////////////////////
// EventLoop
/////////////////////////

static void EventLoop(void)
{
  EventType event;
  Int16     error;
  
  do {
    EvtGetEvent(&event, evtWaitForever);
    if (SysHandleEvent(&event))
      continue;
    if (MenuHandleEvent(CurrentMenu, &event, &error))
      continue;
    if (ApplicationHandleEvent(&event))
      continue;
    FrmDispatchEvent(&event);
  }
  while (event.eType != appStopEvent);
}

/////////////////////////
// PilotMain
/////////////////////////

UInt32  PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
  if (cmd == sysAppLaunchCmdNormalLaunch) {
    if(StartApplication()) {
      EventLoop();
      StopApplication();
    }
  }
  return(0);
}











